@@ -5,13 +5,12 @@ from .wheel_repair import WheelRepair
 from . import patch_dll
 
 
-def subdirectory(s):
-    """Helper for argument parser for validating and normalizing a subdirectory
-    name."""
-    normalized = os.path.normpath(s)
-    if normalized == '.' or normalized == '..' or normalized.startswith('..\\'):
-        raise ValueError(f'invalid subdirectory {s!r}')
-    return normalized
+def subdir_suffix(s: str) -> str:
+    """Helper for argument parser for validating a subdirectory suffix."""
+    if not s or any(c in r'<>:"/\|?*' or ord(c) < 32 for c in s) or \
+            any(s.endswith(x) for x in ('.dist-info', '.data', ' ', '.')):
+        raise argparse.ArgumentTypeError(f'Invalid subdirectory suffix {s!r}')
+    return s
 
 
 def main():
@@ -30,14 +29,15 @@ def main():
     parser_needed = subparsers.add_parser('needed', help=parser_needed_description, description=parser_needed_description)
     for subparser in (parser_show, parser_repair):
         subparser.add_argument('wheel', nargs='+', help='wheel(s) to show or repair')
-        subparser.add_argument('--add-path', default='', help='extra paths(s) to search for DLLs, semicolon-delimited')
-        subparser.add_argument('--add-dll', default='', help='force inclusion of DLL name(s), semicolon-delimited')
-        subparser.add_argument('--no-dll', default='', help='force exclusion of DLL name(s), semicolon-delimited')
+        subparser.add_argument('--add-path', default='', metavar='PATHS', help='extra paths(s) to search for DLLs, semicolon-delimited')
+        subparser.add_argument('--add-dll', default='', metavar='DLLS', help='force inclusion of DLL name(s), semicolon-delimited')
+        subparser.add_argument('--no-dll', default='', metavar='DLLS', help='force exclusion of DLL name(s), semicolon-delimited')
         subparser.add_argument('-v', action='count', default=0, help='verbose mode')
         subparser.add_argument('--extract-dir', help=argparse.SUPPRESS)
     parser_repair.add_argument('-w', '--wheel-dir', dest='target', default='wheelhouse', help='directory to write repaired wheel')
-    parser_repair.add_argument('--no-mangle', default='', help='DLL names(s) not to mangle, semicolon-delimited')
-    parser_repair.add_argument('-L', '--lib-sdir', default='.libs', type=subdirectory, help='subdirectory in package to store vendored DLLs (default .libs)')
+    parser_repair.add_argument('--no-mangle', default='', metavar='DLLS', help='DLL names(s) not to mangle, semicolon-delimited')
+    parser_repair.add_argument('--no-mangle-all', action='store_true', help="don't mangle any DLL names")
+    parser_repair.add_argument('-L', '--lib-sdir', default='.libs', type=subdir_suffix, help='directory suffix in package to store vendored DLLs (default .libs)')
     parser_needed.add_argument('file', help='path to a DLL file')
     args = parser.parse_args()
     if args.command is None:
@@ -61,7 +61,7 @@ def main():
                 wr.show()
             else:  # args.command == 'repair'
                 no_mangles = set(dll_name.lower() for dll_name in args.no_mangle.split(';') if dll_name)
-                wr.repair(args.target, no_mangles, args.lib_sdir)
+                wr.repair(args.target, no_mangles, args.no_mangle_all, args.lib_sdir)
     else:  # args.command == 'needed'
         for dll_name in patch_dll.get_direct_needed(args.file):
             print(dll_name)
