@@ -1,6 +1,7 @@
 """DLL file patching functions."""
 
 import itertools
+import os
 import platform
 import ctypes.util
 import typing
@@ -21,12 +22,18 @@ class PEContext:
         self._pe.close()
 
 
-def get_direct_needed(lib_path: str) -> set:
+def get_direct_needed(lib_path: str, include_delay_imports: bool = True) -> set:
     """Given the path to a shared library, return a set containing the lowercase
-    DLL names of all its direct dependencies."""
+    DLL names of all its direct dependencies.
+    If include_delay_imports is True, delay-loaded dependencies are included.
+    Otherwise, they are not included"""
     with PEContext(lib_path) as pe:
         imports = []
-        for attr in ('DIRECTORY_ENTRY_IMPORT', 'DIRECTORY_ENTRY_DELAY_IMPORT'):
+        if include_delay_imports:
+            attrs = ('DIRECTORY_ENTRY_IMPORT', 'DIRECTORY_ENTRY_DELAY_IMPORT')
+        else:
+            attrs = ('DIRECTORY_ENTRY_IMPORT',)
+        for attr in attrs:
             if hasattr(pe, attr):
                 imports = itertools.chain(imports, getattr(pe, attr))
         needed = set()
@@ -150,7 +157,16 @@ def replace_needed(lib_path: str, old_deps: typing.Iterable, name_map: dict) ->
         return
     with open(lib_path, 'rb') as f:
         buf = f.read()
-    buf = machomachomangler.pe.redll(buf, used_name_map)
+    try:
+        buf = machomachomangler.pe.redll(buf, used_name_map)
+    except ValueError as ex:
+        if "Can't add new section" in str(ex):
+            raise RuntimeError(
+                'Unable to rename the dependencies of '
+                f'{os.path.basename(lib_path)} because this DLL has trailing '
+                'data. If this DLL was created with MinGW, run the strip'
+                'utility. Otherwise, use the --no-mangle flag.')
+        raise ex
     with open(lib_path, 'wb') as f:
         f.write(buf)
     with PEContext(lib_path) as pe:
